@@ -16,6 +16,7 @@ use crate::bindings::{
     settings_write_res_e_SETTINGS_WR_SETTING_REJECTED, settings_write_res_e_SETTINGS_WR_TIMEOUT,
     settings_write_res_e_SETTINGS_WR_VALUE_REJECTED, settings_write_str, size_t,
 };
+use crossbeam_channel::Receiver;
 use crossbeam_utils::thread;
 use log::{debug, error, warn};
 use sbp::{
@@ -23,9 +24,10 @@ use sbp::{
     Sbp, SbpMessage,
 };
 
-use crate::{settings, SettingKind, SettingValue};
+use crate::{Setting, SettingKind, SettingValue};
 
 const SENDER_ID: u16 = 0x42;
+const NUM_WORKERS: usize = 5;
 
 pub struct Client<'a> {
     inner: Box<ClientInner<'a>>,
@@ -40,9 +42,8 @@ struct ClientInner<'a> {
 impl Drop for ClientInner<'_> {
     fn drop(&mut self) {
         unsafe {
-            // Safety: Created via Box::into_raw
-            let _ = Box::from_raw(self.context);
-            // Safety: Created via Box::into_raw
+            // Safety: self.api was created via Box::into_raw.
+            // self.context is freed via the call to settings_destroy
             let _ = Box::from_raw(self.api);
             settings_destroy(&mut self.ctx);
         }
@@ -104,12 +105,43 @@ impl<'a> Client<'a> {
         Client { inner }
     }
 
-    pub fn read_all(&self) -> Vec<Result<ReadByIdxResult, Error<ReadSettingError>>> {
-        const NUM_WORKERS: usize = 5;
+    pub fn read_all(&self) -> (Vec<ReadByIdxResult>, Vec<Error<ReadSettingError>>) {
+        let (_, r) = crossbeam_channel::bounded(0);
+        self.read_all_inner(r)
+    }
 
+    pub fn read_all_timeout(
+        &self,
+        timeout: Duration,
+    ) -> (Vec<ReadByIdxResult>, Vec<Error<ReadSettingError>>) {
+        let (done_s, done_r) = crossbeam_channel::bounded(0);
+        let (stop_s, stop_r) = crossbeam_channel::bounded(NUM_WORKERS);
+        thread::scope(move |scope| {
+            scope.spawn(move |_| {
+                crossbeam_channel::select! {
+                    recv(done_r) -> _ => return,
+                    default(timeout) => (),
+                };
+                warn!("settings read_all timed out");
+                for _ in 0..NUM_WORKERS {
+                    let _ = stop_s.try_send(());
+                }
+            });
+            let res = self.read_all_inner(stop_r);
+            let _ = done_s.send(());
+            res
+        })
+        .unwrap()
+    }
+
+    fn read_all_inner(
+        &self,
+        stop: Receiver<()>,
+    ) -> (Vec<ReadByIdxResult>, Vec<Error<ReadSettingError>>) {
         thread::scope(move |scope| {
             let (idx_s, idx_r) = crossbeam_channel::bounded(NUM_WORKERS);
-            let (results_s, results_r) = crossbeam_channel::unbounded();
+            let (settings_s, settings_r) = crossbeam_channel::unbounded();
+            let (errors_s, errors_r) = crossbeam_channel::unbounded();
 
             scope.spawn(move |_| {
                 let mut idx = 0;
@@ -119,20 +151,34 @@ impl<'a> Client<'a> {
             });
 
             for _ in 0..NUM_WORKERS {
-                let idx_r = idx_r.clone();
-                let results_s = results_s.clone();
-                scope.spawn(move |_| {
-                    for idx in idx_r.iter() {
-                        match self.read_by_index(idx).map_err(Error::Err).transpose() {
-                            Some(res) => results_s.send(res).expect("results channel closed"),
-                            None => break,
+                scope.spawn({
+                    let idx_r = idx_r.clone();
+                    let settings_s = settings_s.clone();
+                    let errors_s = errors_s.clone();
+                    let stop = stop.clone();
+                    move |_| {
+                        for idx in idx_r.iter() {
+                            if stop.try_recv().is_ok() {
+                                break;
+                            }
+                            match self.read_by_index(idx) {
+                                Ok(Some(setting)) => settings_s.send(setting).unwrap(),
+                                Ok(None) => break,
+                                Err(err) => {
+                                    errors_s.send(Error::Err(err)).unwrap();
+                                    break;
+                                }
+                            }
                         }
                     }
                 });
             }
-
-            drop(results_s);
-            results_r.iter().collect::<Vec<_>>()
+            // we need to drop these before calling `.iter()` to disconnect the channels
+            // otherwise `.iter()` will never terminate
+            drop(idx_r);
+            drop(settings_s);
+            drop(errors_s);
+            (settings_r.iter().collect(), errors_r.iter().collect())
         })
         .expect("read_all worker thread panicked")
     }
@@ -195,9 +241,7 @@ impl<'a> Client<'a> {
         group: &str,
         name: &str,
     ) -> Option<Result<SettingValue, Error<ReadSettingError>>> {
-        let setting = settings()
-            .iter()
-            .find(|s| s.group == group && s.name == name)?;
+        let setting = Setting::find(group, name)?;
 
         let cgroup = match CString::new(group) {
             Ok(cgroup) => cgroup,
@@ -724,7 +768,7 @@ mod tests {
         mut wtr: impl Write + 'static,
         group: &str,
         name: &str,
-    ) -> Option<Result<settings::SettingValue, Error<ReadSettingError>>> {
+    ) -> Option<Result<SettingValue, Error<ReadSettingError>>> {
         scope(move |scope| {
             let source = LinkSource::new();
             let link = source.link();
@@ -749,14 +793,14 @@ mod tests {
 
         let request_msg = MsgSettingsReadReq {
             sender_id: Some(SETTINGS_SENDER_ID),
-            setting: SbpString::from(format!("{}\0{}\0", group, name).to_string()),
+            setting: SbpString::from(format!("{}\0{}\0", group, name)),
         };
 
         stream.wait_for(sbp::to_vec(&request_msg).unwrap().as_ref());
 
         let reply_msg = MsgSettingsReadResp {
             sender_id: Some(0x42),
-            setting: SbpString::from(format!("{}\0{}\010\0", group, name).to_string()),
+            setting: SbpString::from(format!("{}\0{}\010\0", group, name)),
         };
 
         stream.push_bytes_to_read(sbp::to_vec(&reply_msg).unwrap().as_ref());
@@ -773,14 +817,14 @@ mod tests {
 
         let request_msg = MsgSettingsReadReq {
             sender_id: Some(SETTINGS_SENDER_ID),
-            setting: SbpString::from(format!("{}\0{}\0", group, name).to_string()),
+            setting: SbpString::from(format!("{}\0{}\0", group, name)),
         };
 
         stream.wait_for(sbp::to_vec(&request_msg).unwrap().as_ref());
 
         let reply_msg = MsgSettingsReadResp {
             sender_id: Some(0x42),
-            setting: SbpString::from(format!("{}\0{}\0True\0", group, name).to_string()),
+            setting: SbpString::from(format!("{}\0{}\0True\0", group, name)),
         };
 
         stream.push_bytes_to_read(sbp::to_vec(&reply_msg).unwrap().as_ref());
@@ -797,14 +841,14 @@ mod tests {
 
         let request_msg = MsgSettingsReadReq {
             sender_id: Some(SETTINGS_SENDER_ID),
-            setting: SbpString::from(format!("{}\0{}\0", group, name).to_string()),
+            setting: SbpString::from(format!("{}\0{}\0", group, name)),
         };
 
         stream.wait_for(sbp::to_vec(&request_msg).unwrap().as_ref());
 
         let reply_msg = MsgSettingsReadResp {
             sender_id: Some(SETTINGS_SENDER_ID),
-            setting: SbpString::from(format!("{}\0{}\00.1\0", group, name).to_string()),
+            setting: SbpString::from(format!("{}\0{}\00.1\0", group, name)),
         };
 
         stream.push_bytes_to_read(sbp::to_vec(&reply_msg).unwrap().as_ref());
@@ -822,14 +866,14 @@ mod tests {
 
         let request_msg = MsgSettingsReadReq {
             sender_id: Some(0x42),
-            setting: SbpString::from(format!("{}\0{}\0", group, name).to_string()),
+            setting: SbpString::from(format!("{}\0{}\0", group, name)),
         };
 
         stream.wait_for(sbp::to_vec(&request_msg).unwrap().as_ref());
 
         let reply_msg = MsgSettingsReadResp {
             sender_id: Some(SETTINGS_SENDER_ID),
-            setting: SbpString::from(format!("{}\0{}\0foo\0", group, name).to_string()),
+            setting: SbpString::from(format!("{}\0{}\0foo\0", group, name)),
         };
 
         stream.push_bytes_to_read(sbp::to_vec(&reply_msg).unwrap().as_ref());
@@ -847,14 +891,14 @@ mod tests {
 
         let request_msg = MsgSettingsReadReq {
             sender_id: Some(SETTINGS_SENDER_ID),
-            setting: SbpString::from(format!("{}\0{}\0", group, name).to_string()),
+            setting: SbpString::from(format!("{}\0{}\0", group, name)),
         };
 
         stream.wait_for(sbp::to_vec(&request_msg).unwrap().as_ref());
 
         let reply_msg = MsgSettingsReadResp {
             sender_id: Some(0x42),
-            setting: SbpString::from(format!("{}\0{}\0Secondary\0", group, name).to_string()),
+            setting: SbpString::from(format!("{}\0{}\0Secondary\0", group, name)),
         };
 
         stream.push_bytes_to_read(sbp::to_vec(&reply_msg).unwrap().as_ref());
